require "defines"
require "utils"

game.oninit(function() OnInit() end)
game.onload(function() OnLoad() end)
game.onevent(defines.events.ontick, function(event) OnTick(event) end)
game.onevent(defines.events.onbuiltentity, function(event) OnBuiltEntity(event) end)
game.onevent(defines.events.onentitydied, function(event) OnEntityDied(event) end)
game.onevent(defines.events.onpreplayermineditem, function(event) OnPrePlayerMinedItem(event) end)

function OnInit()
	if glob.sb == nil then
		glob.sb = {	version = utils.currentVersion,
					switchers = {}
				  }
	end
	if glob.sb.version ~= utils.currentVersion then
		game.player.print("Migrating BeltSwitchers from version " .. glob.sb.version .. " to version " .. utils.currentVersion .. ".")
		utils.migrations[glob.sb.version]()
		game.player.print("Migration successful.")
	end
end

function OnLoad()
	-- Call OnInit to ensure glob.sb exists.
	OnInit()
end
						  
function OnTick(_Event)
	for i, switcher in pairs(glob.sb.switchers) do
		local scanArea = GetSwitchScanArea(switcher.ent.direction, switcher.ent.position)
		
		scan = {left = game.findentitiesfiltered{	name = "item-on-ground",
							  				 		area = scanArea.left
												},
				right = game.findentitiesfiltered{	name = "item-on-ground",
							  					 	area = scanArea.right
												 }
			   }
		
		--if #scan.left > 1 or #scan.right > 1 then
			--game.player.print("BeltSwitchers ERROR: Found (" .. #scan.left .. ", " .. #scan.right .. ") entities")
		--end
		if switcher.stuckStack.left and switcher.stuckStack.left.valid then
			MoveStackToRight(switcher.stuckStack.left, switcher)
		elseif #scan.left == 1 then
			MoveStackToRight(scan.left[1], switcher)
		end
		if switcher.stuckStack.right and switcher.stuckStack.right.valid then
			MoveStackToLeft(switcher.stuckStack.right, switcher)
		elseif #scan.right == 1 then
			MoveStackToLeft(scan.right[1], switcher)
		end
	end
end

function OnBuiltEntity(_Event)
	if not IsSwitcher(_Event.createdentity) then return end
	
	AddSwitcher(_Event.createdentity)
end

function OnEntityDied(_Event)
	if not IsSwitcher(_Event.entity) then return end
	
	RemoveSwitcher(_Event.entity)
end

function OnPrePlayerMinedItem(_Event)
	if not IsSwitcher(_Event.entity) then return end
	
	RemoveSwitcher(_Event.entity)
end

local VV = 0.25
local targetOff = {[0] = {left = {x = -0.23, y = -VV  }, right = {x =  0.23, y = -VV  }},
				   [2] = {left = {x =  VV  , y = -0.23}, right = {x =  VV  , y =  0.23}},
				   [4] = {left = {x =  0.23, y =  VV  }, right = {x = -0.23, y =  VV  }},
				   [6] = {left = {x = -VV  , y =  0.23}, right = {x = -VV  , y = -0.23}}
						 }

function MoveStackToRight(_Stack, _Switcher)
	newPos = {x = _Switcher.ent.position.x + targetOff[_Switcher.ent.direction].right.x,
			  y = _Switcher.ent.position.y + targetOff[_Switcher.ent.direction].right.y}
	local moved, stuckStack = MoveStack(_Stack, newPos)
	if not moved and stuckStack and stuckStack.valid then
		_Switcher.stuckStack.left = stuckStack
	end
end

function MoveStackToLeft(_Stack, _Switcher)
	newPos = {x = _Switcher.ent.position.x + targetOff[_Switcher.ent.direction].left.x,
			  y = _Switcher.ent.position.y + targetOff[_Switcher.ent.direction].left.y}
	local moved, stuckStack = MoveStack(_Stack, newPos)
	if not moved and stuckStack and stuckStack.valid then
		_Switcher.stuckStack.right = stuckStack
	end
end

function MoveStack(_Stack, _NewPos)
	-- Check if the target location is empty.
	local possiblePos = game.findnoncollidingposition("item-on-ground", _NewPos, 0.10, 0.01)
	if possiblePos then
		game.createentity{position=possiblePos, name="item-on-ground", stack={name=_Stack.stack.name, count=_Stack.stack.count}}
		_Stack.destroy()
											
		return true, nil
	else
		-- Teleport the stack, locking it in position.
		_Stack.teleport(_Stack.position)
		return false, _Stack
	end
	return false, nil
end

function IsSwitcher(_Entity)
	if not _Entity or not _Entity.valid then
		return false
	end
	if _Entity.name == "beltswitcher" then
		return true
	end
	return false
end

function AddSwitcher(_Entity)
	table.insert(glob.sb.switchers, {	ent = _Entity,
										stuckStack = {left = nil, right = nil}
									}
				)
end

function RemoveSwitcher(_Entity)
	for i, switcher in pairs(glob.sb.switchers) do
		if switcher.ent.equals(_Entity) then
			table.remove(glob.sb.switchers, i)
			break
		end
	end
end
function GetSwitchScanArea(_Direction, _Position)
	local offL = 0.05
	local offH = 0.3
	if _Direction == 0 then
		return {left  = {{_Position.x + -offH, _Position.y        }, {_Position.x       , _Position.y + offL}},
				right = {{_Position.x        , _Position.y        }, {_Position.x + offH, _Position.y + offL}}
			   }
	elseif _Direction == 2 then
		return {left  = {{_Position.x + -offL, _Position.y + -offH}, {_Position.x       , _Position.y       }},
				right = {{_Position.x + -offL, _Position.y        }, {_Position.x       , _Position.y + offH}}
			   }
	elseif _Direction == 4 then
		return {left  = {{_Position.x        , _Position.y + -offL}, {_Position.x + offH, _Position.y       }},
				right = {{_Position.x + -offH, _Position.y + -offL}, {_Position.x       , _Position.y       }}
			   }
	else
		return {left  = {{_Position.x        , _Position.y        }, {_Position.x + offL, _Position.y + offH}},
				right = {{_Position.x        , _Position.y + -offH}, {_Position.x + offL, _Position.y       }}
			   }
	end
end